"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const fs_1 = require("fs");
const net_1 = require("net");
const web3_types_1 = require("web3-types");
const web3_errors_1 = require("web3-errors");
const web3_utils_1 = require("web3-utils");
class IpcProvider extends web3_types_1.Web3BaseProvider {
    constructor(socketPath) {
        super();
        this._emitter = new events_1.EventEmitter();
        this._connectionStatus = 'disconnected';
        this._socketPath = socketPath;
        this._socket = new net_1.Socket();
        this._requestQueue = new Map();
        this.connect();
        this.waitOptions = {
            timeOutTime: 5000,
            maxNumberOfAttempts: 10,
        };
        this.chunkResponseParser = new web3_utils_1.ChunkResponseParser();
        this.chunkResponseParser.onError(() => {
            this._clearQueues();
        });
    }
    getStatus() {
        return this._connectionStatus;
    }
    supportsSubscriptions() {
        return true;
    }
    on(type, callback) {
        this._emitter.on(type, callback);
    }
    once(type, callback) {
        this._emitter.once(type, callback);
    }
    removeListener(type, callback) {
        this._emitter.removeListener(type, callback);
    }
    connect() {
        if (!(0, fs_1.existsSync)(this._socketPath)) {
            throw new web3_errors_1.InvalidClientError(this._socketPath);
        }
        try {
            this._addSocketListeners();
            this._socket.connect({ path: this._socketPath });
        }
        catch (e) {
            throw new web3_errors_1.InvalidConnectionError(this._socketPath);
        }
    }
    disconnect() {
        this._requestQueue.clear();
        this._removeSocketListeners();
        this._socket.end();
    }
    reset() {
        this._requestQueue.clear();
        this._removeSocketListeners();
        this._addSocketListeners();
    }
    get waitTimeOut() {
        return this.waitOptions.timeOutTime;
    }
    set waitTimeOut(timeOut) {
        this.waitOptions.timeOutTime = timeOut;
    }
    get waitMaxNumberOfAttempts() {
        return this.waitOptions.maxNumberOfAttempts;
    }
    set waitMaxNumberOfAttempts(maxNumberOfAttempts) {
        this.waitOptions.maxNumberOfAttempts = maxNumberOfAttempts;
    }
    waitForConnection() {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => {
                let currentAttempt = 0;
                const interval = setInterval(() => {
                    if (currentAttempt > this.waitMaxNumberOfAttempts - 1) {
                        clearInterval(interval);
                        reject(new web3_errors_1.ConnectionNotOpenError());
                    }
                    else if (this.getStatus() === 'connected') {
                        clearInterval(interval);
                        resolve();
                    }
                    currentAttempt += 1;
                }, this.waitTimeOut / this.waitMaxNumberOfAttempts);
            });
        });
    }
    request(request) {
        return __awaiter(this, void 0, void 0, function* () {
            if ((0, web3_utils_1.isNullish)(this._socket))
                throw new Error('IPC connection is undefined');
            const requestId = web3_utils_1.jsonRpc.isBatchRequest(request) ? request[0].id : request.id;
            if ((0, web3_utils_1.isNullish)(requestId))
                throw new Error('Request Id not defined');
            if (this.getStatus() !== 'connected') {
                yield this.waitForConnection();
            }
            if (!this._socket.writable) {
                console.error('Can not send a request. The internal socket is not `writable`. Request data: ', request);
                const dummyPromise = new web3_utils_1.Web3DeferredPromise();
                return dummyPromise;
            }
            try {
                const defPromise = new web3_utils_1.Web3DeferredPromise();
                this._requestQueue.set(requestId, defPromise);
                this._socket.write(JSON.stringify(request));
                return defPromise;
            }
            catch (error) {
                this._requestQueue.delete(requestId);
                throw error;
            }
        });
    }
    removeAllListeners(type) {
        this._emitter.removeAllListeners(type);
    }
    _onMessage(e) {
        const responses = this.chunkResponseParser.parseResponse(typeof e === 'string' ? e : e.toString('utf8'));
        if (!responses) {
            return;
        }
        for (const response of responses) {
            if (web3_utils_1.jsonRpc.isResponseWithNotification(response) &&
                response.method.endsWith('_subscription')) {
                this._emitter.emit('message', undefined, response);
                return;
            }
            const requestId = web3_utils_1.jsonRpc.isBatchResponse(response) ? response[0].id : response.id;
            const requestItem = this._requestQueue.get(requestId);
            if (!requestItem) {
                return;
            }
            if (web3_utils_1.jsonRpc.isBatchResponse(response) || web3_utils_1.jsonRpc.isResponseWithResult(response)) {
                this._emitter.emit('message', undefined, response);
                requestItem.resolve(response);
            }
            else {
                this._emitter.emit('message', response, undefined);
                requestItem === null || requestItem === void 0 ? void 0 : requestItem.reject(new web3_errors_1.ResponseError(response));
            }
            this._requestQueue.delete(requestId);
        }
    }
    _clearQueues(event) {
        if (this._requestQueue.size > 0) {
            this._requestQueue.forEach((request, key) => {
                request.reject(new web3_errors_1.ConnectionNotOpenError(event));
                this._requestQueue.delete(key);
            });
        }
        this._removeSocketListeners();
    }
    _onConnect() {
        this._connectionStatus = 'connected';
        this._emitter.emit('connect');
    }
    _onDisconnect() {
        this._connectionStatus = 'disconnected';
        this._emitter.emit('disconnect');
    }
    _onClose(event) {
        this._clearQueues(event);
        this._removeSocketListeners();
    }
    _addSocketListeners() {
        this._socket.on('connect', this._onConnect.bind(this));
        this._socket.on('end', this._onDisconnect.bind(this));
        this._socket.on('close', this._onClose.bind(this));
        this._socket.on('data', this._onMessage.bind(this));
    }
    _removeSocketListeners() {
        var _a, _b, _c, _d;
        (_a = this._socket) === null || _a === void 0 ? void 0 : _a.removeAllListeners('connect');
        (_b = this._socket) === null || _b === void 0 ? void 0 : _b.removeAllListeners('end');
        (_c = this._socket) === null || _c === void 0 ? void 0 : _c.removeAllListeners('close');
        (_d = this._socket) === null || _d === void 0 ? void 0 : _d.removeAllListeners('data');
    }
}
exports.default = IpcProvider;
//# sourceMappingURL=index.js.map